/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                             Page Layout                                       *
*********************************************************************************
*   $Id: FXPageFormat.h,v 1.2 2000/11/01 11:27:10 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXPageFormat.h,v $
* Revision 1.2  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.1.1.1  2000/10/29 14:17:49  tekno
* Initial release
*
*
*
********************************************************************************/
#ifndef FXPAGEFORMAT_H
#define FXPAGEFORMAT_H

class FXTextAnchor;
class FXTextBracket;

#ifndef FXROWINFO_H
#include "FXRowInfo.h"
#endif
#ifndef FXARRAY_H
#include "FXElement.h"
#include "FXArray.h"
#endif



/**
* Page formatter.  Interface object between FXPLText and FXTextSrc.
* FXPageFormat is basic class which represents each line as a single row, truncating on
* the right to fit the display.
* FXWrapPageFormat is derived from the above.  It splits each line into one or more
* rows, using a break-character heuristic.  Default break character is one satisfying
* isspace().  Can also break at the last column displayable <- FIXME.
* All classes have member functions for FXPLText to query the current character/attribute
* matrix.  They also forward and translate messages between FXPLText and FXTextSrc.
* Member functions translate between row- and line-based coordinates, as well as 'pos' i.e.
* character offsets in the text source.  Page formatter also handles text which is represented
* by more than 1 position in matrix, such as tabs and control characters.
*
* Note that we talk about 'absolute' and 'relative' row numbers.  Absolute row numbers are
* counted from the start of the text source, whereas relative rows are counted from the
* first row represented in the character matrix.  The relation is abs_row = rel_row + toprow.
*
*/
class FXAPI FXPageFormat : public FXObject {
  FXDECLARE(FXPageFormat)
protected:
  FXbool ok;		// Flag to indicate construction completed successfully
  FXPLText *    pltext;	// Display widget
  FXTextSrc *	textsrc;// Source of line buffer contents
  FXbool	textsrc_private; // Whether the above is unique to this object
  FXchar *	matrix;	// The character matrix.  Dimensioned cols*rows.
  FXattr *	attrs;	// The attribute matrix.  Dimensioned cols*rows.  One position for each
  			// char in matrix.
  FXchar *	matrow;	// Single row buffer for matrix.  Dimensioned cols.
  FXattr *	attrow;	// Single row buffer for attributes.  Dimensioned cols.
  FXint	cols;	// Columns per row of the above
  FXRowInfo *	rowvec; // Row information.  Dimensioned rows plus a few either side.
  FXint	rows;	// Rows in the above.
  
  FXint	toprow;	// Absolute row number of first row in matrix - 'scroll position'.
  
  FXint	flags;	// Options
  FXint	tabcols;// Tab columns.
  
  FXbool	defer;	// If true, defer calling pltext->updateDirty() because more changes
  			// are coming.  This is set during refresh() processing.
  FXbool	wasmore;// If true, then change() has been called with more true.
  FXArray<FXAnchorInfo> renderables;
  	// List of anchors (and brackets, if they have glyphs) which are visible in current
  	// page.  The widget can iterate through these to speed up the redraw process
  FXArray<FXAnchorInfo> newrend;
  	// Temporary updated list of renderable anchors.  This is updated during multiple
  	// anchor changes when defer is true.  At the end of the update process, this is
  	// compared with the previous settings (in renderables) so that the most efficient
  	// graphical update may be performed.
  FXbool	wasanch;// If true, at least one update was made to newrend, thus renderables
  			// needs to be updated and any graphics performed.
  
protected:
  FXPageFormat();
  
  // Given character c positioned in display column col, return number of display positions
  // taken by the visual representation of c, and increment col by the same.
  inline FXint expand(FXint & col, FXchar c) const;
  
  // Given character c positioned in display column col, return number of display positions
  // taken by the visual representation of c, and increment col by the same.  In addition,
  // the buffer at ptr (with length len) is updated with the visual representation (e.g. for
  // control chars, they are written as two chars e.g. ^A) at offset col (and beyond).
  // attr points to the corresponding attribute buffer (of same length).  This can be
  // updated as well to indicate pseudo-chars.   
  FXint expand(FXint & col, FXchar c, FXchar * ptr, FXint len, FXattr * attr);
  
  // Return the number of characters to the next end-of-row, starting from position pos.
  // pos is updated to the position of the start of the next row.  If the range of
  // characters implied by this function includes a newline, then that newline will be
  // the last character of the row.  The initial position must be the start of
  // a row, but this may not be verified.  If the row happens to be ended by a newline,
  // then linect must be incremented by 1.
  virtual FXint nextSplit(FXint & pos, FXint & linect) const;
  
  // Assuming pos is the start of a row, return the apparent column number of the char
  // at offset offs from pos.  This is necessary for handling tabs and other chars which expand
  // to more than one display position.  It converts from char positions to display
  // coordinates.
  virtual FXint offs2col(FXint offs, FXint pos) const;
  
  // Assuming pos is the start of a row, return the offset (i.e. beyond pos) of the char
  // in apparent column number col.  This is for converting display coordinates to char
  // positions.  len is assumed to be the length of the row.
  virtual FXint col2offs(FXint col, FXint pos, FXint len) const;
  
  // Set the characters of specified absolute row of matrix, starting with position pos.
  void setRowChars(FXint r, FXint pos, FXint len);
  
  // Called whenever number of columns changes, in order to allow row<->line mapping
  // to be updated.  This should not look at current matrix members, since
  // they may not have been updated yet.  After this is called, getLineOfRow() and
  // getRowOfLine() should assume that the requested column change is in effect.
  // This function must update cols.
  virtual void changeCols(FXint newcols);
  
  // Called whenever text source lines change.  This is invoked from change() [public member]
  // for the purpose of updating row<->line mapping.  Must return the change (+ or -) in
  // the number of rows.
  // startline and endline are set to the first and last lines actually changed, but this
  // function can widen the range if it needs to operate on a coarser granularity.  startpos
  // is the first char affected, endpos is the last+1, and deltachars is the nett number
  // of chars inserted.
  virtual FXint changeLines(FXint startline, FXint endline, FXint deltalines,
                            FXint startpos, FXint endpos, FXint deltachars);
  
private:
  FXPageFormat(const FXPageFormat&);
  FXPageFormat& operator=(const FXPageFormat&);
  
  // (re-)allocate space for matrix and row info vector.
  FXint makeMatrixAndVector(FXint r, FXint c);

protected:
  // Notify pltext object of some change that it might be interested in.
  long notifyPLText(FXint id) 
    { return pltext ? pltext->handle(this, MKUINT(id, SEL_COMMAND), NULL) : 0; }
  
  void updateRenderables();
  FXint removeRenderable(const FXTextAnchor * ta, FXbool temp);
  FXint addRenderable(const FXTextAnchor * ta, 
     FXshort r1, FXshort c1, FXshort r2, FXshort c2, FXbool temp);
  void dirtyAnchor(const FXAnchorInfo & ai, FXint where); 
  
public:
  /// Construct, linking to a display widget (FXPLText) and a text source (FXTextSrc).
  /// plt must be non-null.  If ds is null, a default in-memory text source is constructed for
  /// exclusive use of this page format.  New object adds itself to list in text source.
  /// r and c specify an initial display size.
  FXPageFormat(FXPLText * plt, FXTextSrc * ds = NULL, FXint opts = 0, FXint r = 0, FXint c = 0);
  
  /// Test whether constructor completed successfully.  Since ctors may allocate memory etc.,
  /// it is necessary for the caller to check whether construction was sucessful.  If not,
  /// then the dtor should be invoked as soon as possible and an error flagged to the user.
  /// Note that the ctor is exclusively responsible for setting this flag.  After construction,
  /// this flag can indicate an error encountered during change() processing.
  FXbool isOk() const { return ok; }
  
  /// Initial setup or refresh of matrix when ready to display data.
  void refresh();
  
  /// Return line number and row within line, given an absolute row number.
  virtual FXint getLineOfRow(FXint r, FXint & linerow) const;
  
  /// Return absolute row number and number of rows in line, given line number.
  virtual FXint getRowOfLine(FXint line, FXint & numrows) const;
  
  /// Return total number of rows.
  virtual FXint numRows() const;
  
  /// Return character position of first char in given row.
  virtual FXint getPosOfRow(FXint r) const;
  
  /// Return row number of row containing the given position.  Also returns position of 
  /// first char in that row.  atend has same meaning as in getMatrixCoordsOfPos().
  virtual FXint getRowOfPos(FXint pos, FXint & startpos, FXbool atend=FALSE) const;
  
  /// Return row number of row containing the given position.
  virtual FXint getRowOfPos(FXint pos, FXbool atend=FALSE) const;
  
  /// Return number of character units taken up by the first numchars characters in the
  /// given row.  This may be >= numchars, since tabs may expand to several positions.
  virtual FXint rowWidth(FXint r, FXint numchars) const;
  
  /// Return the absolute char position of the last character which is completed before the
  /// given number of character units after the start of the given row.  This is the
  /// complementary function to rowWidth().  If include_nl is false, then the terminating
  /// newline (if any) is not counted.  This function only considers characters in the
  /// given row, not any following rows.
  virtual FXint rowPos(FXint r, FXint numcharspaces, FXbool include_nl) const;
  
  /// Update state in readiness for text source 'replace' operation.  startline is first line
  /// replaced by change.  endline is last line (counting in 'old' buffer) replaced.  deltalines
  /// is the nett number of lines to be added.  [startpos, endpos) and deltachars
  /// indicate the same change, in character position terms.  This is called by the text source
  /// immediately prior to change(), and is used to save the current state (e.g. row mapping).
  /// The parameters which are passed to this function are also passed to change(). 
  virtual void preChange(FXint startline, FXint endline, FXint deltalines,
               FXint startpos, FXint endpos, FXint deltachars);

  
  /// Return matrix row and column for given position.  Returns 0 if in matrix, or -1
  /// if out of bounds low side or +1 if out of bounds high side.  If column is outside
  /// bounds, it is set back to last column of row.  If row is out of bounds of current
  /// matrix, returns a negative row (for before start), or a row greater or equal to
  /// number of rows in matrix (if past end).  In the case that a position is ambiguous
  /// because it is at the end of a row which is not the last row of a line, the atend
  /// parameter is used to specify a preference: at the end of the row or the start of
  /// the next row. 
  FXint getMatrixCoordsOfPos(FXint pos, FXint & r, FXint & c, FXbool atend) const;
  
  /// Return position of first character in row which contains pos. 
  FXint rowStart(FXint pos) const { return getPosOfRow(getRowOfPos(pos)); }
  
  /// Return position of last character in row which contains pos. 
  FXint rowEnd(FXint pos) const { return getPosOfRow(getRowOfPos(pos)+1)-1; } 
  
  /// Return position of first character in row after the row which contains pos. 
  FXint nextRow(FXint pos) const { return getPosOfRow(getRowOfPos(pos)+1); }
  
  /// Return position of first character in row before the row which contains pos. 
  FXint prevRow(FXint pos) const { return getPosOfRow(getRowOfPos(pos)-1); }
  
  /// Resize the matrix to specified rows and columns.  Returns NZ if error.  If the number of
  /// columns does not change, the current top row remains at top.  Otherwise, the line of the
  /// top row has its first row placed at top.  Notifies pltext.
  FXint resize(FXint r, FXint c);
  
  /// Update state because of text source 'replace' operation.  Notifies pltext.  This
  /// is called by the text source immediately after preChange(), and after the text source
  /// has updated the buffer and its line mapping.  This may also be called
  /// when something requires recomputation of row breaks for reasons other than resizing
  /// the number of columns.  If more is true, then more updates are to follow; graphical
  /// update (notification to pltext) should be deferred until called with more false.
  /// Parameters other than 'more' are the same as passed to preChange().  Return value is
  /// 0 for OK, or non-zero if error, in which case isOK() will return false.
  FXint change(FXint startline, FXint endline, FXint deltalines,
               FXint startpos, FXint endpos, FXint deltachars, FXbool more = FALSE);
  
  /// Called from the text source when an anchor or derivative is deleted, created,
  /// or updated.  This allows the attribute matrix to be updated appropriately, and the
  /// pltext widget to be notified so that it can re-render or undraw the glyph or chars.
  /// 'on' is true if the marker is set, otherwise it is being removed.
  /// If more is true, then graphical update should be deferred since there are more
  /// marks being updated at the same time.
  FXint markUpdated(const FXTextAnchor * ta, FXint sp, FXint ep, FXbool on, FXbool more); 
  
  /// Return character/attribute at given absolute row/column.  Returns non-zero for
  /// r/c which is out of current matrix bounds, else returns zero and sets ch and attr.
  FXint getCharAttr(FXint r, FXint c, FXchar & ch, FXattr & attr) const;
  
  /// Return pointer to pltext.  May be null if none currently set.
  FXPLText * getPLText() const { return pltext; } 
  
  /// Change pltext widget.  Returns previous one, or null.  New one may be null.
  /// This should only be called by a pltext member. 
  FXPLText * setPLText(FXPLText * plt);
  
  /// Return pointer to text source.  Text source cannot be changed; it is specified only
  /// to constructor.
  FXTextSrc * getTextSrc() const { return textsrc; }
  
  /// Return the character matrix.  Caller must not modify the contents, and the returned
  /// pointer is only valid while the dimensions do not change.
  const FXchar * getMatrix() const { return matrix; }
  
  /// Return attribute matrix under same conditions as getMatrix().
  const FXattr * getAttrs() const { return attrs; }
  
  /// Return the row info vector.  Caller must not modify the contents, and the returned
  /// pointer is only valid while the dimensions do not change.
  const FXRowInfo * getRowVector() const { return rowvec; }
  
  /// Set the top matrix row to given absolute row.  Effectively scrolls the contents.
  /// Notifies pltext.
  void setTopRow(FXint r);
  
  /// Return current top absolute row number.
  FXint getTopRow() const { return toprow; }
  
  /// Return current matrix number of rows (vertical dimension).
  FXint getRows() const { return rows; }
  
  /// Return current matrix number of columns (horizontal dimension).
  FXint getCols() const { return cols; }
  
  /// Return line number of top absolute row number.
  FXint getFirstLine() const;
  
  /// Return line number of bottom absolute row number.
  FXint getLastLine() const;
  
  /// Return char position of first char displayed. 
  FXint getFirstPos() const;
  
  /// Return char position of last char displayed.
  FXint getLastPos() const;
  
  /// Return number of anchors which are visible in this pager, and have a glyph (i.e.
  /// don't count selections etc which only highlight the text differently).
  FXint getNumRenderables() const { return renderables.no(); }
  
  /// Return pointer to renderable anchor, given index in range [0..getNumRenderables)
  const FXAnchorInfo * getRenderable(FXint idx) const { return &renderables[idx]; } 
  
  /// Return tab columns.
  FXint getTabColumns() const { return tabcols; }
  
  /// Set tab columns.
  void setTabColumns(FXint t);
  
  /// Set or clear shared text source status
  void setSharedTextSrc(FXbool share = TRUE) { textsrc_private = !share; }
  
  /// Save to a stream
  virtual void save(FXStream& store) const;

  /// Load from a stream
  virtual void load(FXStream& store);

  /// Destructor.  Unlinks itself from data target list.
  virtual ~FXPageFormat();
  };


class FXTextMap;

class FXAPI FXWrapPageFormat : public FXPageFormat {
  FXDECLARE(FXWrapPageFormat)
protected:
  FXCharset	delimiters;		// Line-splitting delimiters.
  FXuchar	delimtab[256];		// Same as above, except one byte per element.
  FXTextMap	* map;			// Mapping of rows (Pn) to lines (Ln).

  FXWrapPageFormat(); 
  
  virtual FXint nextSplit(FXint & pos, FXint & linect) const;
  virtual void changeCols(FXint newcols);
  virtual FXint changeLines(FXint startline, FXint endline, FXint deltalines,
                            FXint startpos, FXint endpos, FXint deltachars);
  
  // Reload the line break mapping
  void loadMap();
  
  // Reload the delimtab from delimiters.
  void loadDelimtab(); 
  
  FXint getRowOfPos(FXint pos, FXint & node, FXint & Ln, FXint & Pn) const;
  
  // Variables used to pass data from preChange() to changeLines().
  FXint row1, node1, Ln1, Pn1, row2, node2, Ln2, Pn2;
  FXint lp1, lp2;
  
  
private:
  FXWrapPageFormat(const FXWrapPageFormat&);
  FXWrapPageFormat& operator=(const FXWrapPageFormat&);

public:

  FXWrapPageFormat(FXPLText * plt, FXTextSrc * ds = NULL, FXint opts = 0, 
                   FXint r = 0, FXint c = 0, const FXchar * delimset = " \t~.,/\\!%^&*(-=+{|[:;<>?");
  virtual FXint getLineOfRow(FXint r, FXint & linerow) const;
  virtual FXint getRowOfLine(FXint line, FXint & numrows) const;
  virtual FXint numRows() const;
  virtual FXint getPosOfRow(FXint r) const;
  virtual FXint getRowOfPos(FXint pos, FXint & startpos, FXbool atend=FALSE) const;
  virtual FXint getRowOfPos(FXint pos, FXbool atend=FALSE) const;
  virtual void preChange(FXint startline, FXint endline, FXint deltalines,
               FXint startpos, FXint endpos, FXint deltachars);
  
  /// Set new line-splitting delimiters.  If empty charset, then lines will be split into 
  /// fixed width rows.
  void setDelimiters(const FXCharset & delim); 
  
  /// Return current delimiters (may be null).
  const FXCharset & getDelimiters() const { return delimiters; }
  
  virtual void save(FXStream& store) const;
  virtual void load(FXStream& store);
  virtual ~FXWrapPageFormat();
  };

#endif