/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                        Anchors, brackets and cursors                          *
*********************************************************************************
*   $Id: FXTextAnchor.h,v 1.2 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXTextAnchor.h,v $
* Revision 1.2  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.1.1.1  2000/10/29 14:19:09  tekno
* Initial release
*
*
*
********************************************************************************/
#ifndef FXTEXTANCHOR_H
#define FXTEXTANCHOR_H
 
class FXApp;
class FXPLText;
class FXPageFormat;
class FXTextSrc;

/** Pointer to this struct is passed as the message parameter of ID_REPLACE messages
*   sent to a text anchor (or, usually, bracket).  The parameters are similar in
*   semantics, name and order to those of FXTextSrc::replace().  An additional parameter,
*   from_widget, indicates whether the replacement originates by the user manipulating
*   an FXPLText widget, or from some other process (NULL).
*/

struct FXTextReplacement {
  FXint startpos;	/// Position of what to replace
  FXint endpos;		/// Last position (+1) of what to replace
  const FXchar * text;	/// Data (may be NULL if n==0)
  FXint n;		/// Length of above string
  FXbool defer;		/// TRUE if not last in a sequence of replacements - defers GUI update.
  const FXPLText * from_widget;	/// Non-null if user input from widget, else from application.
  };

/** Base class for FXTextAnchor, which represents a position in the FXTextSrc.  FXTextBracket
*   is derived from this, and represents a range of zero or more characters.
*/

class FXAPI FXTextAnchor : public FXObject {
  FXDECLARE(FXTextAnchor)
protected:
  FXTextSrc * owner;	// Text source owner
  FXuint attrs;		// Attributes
  FXint pos;		// Starting position
  FXint epos;		// Ending position (==pos for base class)
  FXbool visible;	// False: not visible in any; True: may be visible
  enum {
    MUTEX_CURSOR = 1<<0	// Used by text cursors
   ,MUTEX_LAST = 1<<1
    };
  
public:
  void * data;		/// User-defined data
  
  // SEL_COMMAND/ID_REPLACE message
  long onCmdReplace(FXObject*,FXSelector,void*);
  // SEL_IO_READ/ID_SOCKET message from an FXInputHandle.
  long onIORead(FXObject*,FXSelector,void*);
  
  enum {
    ID_REPLACE,
    ID_SOCKET,
    ID_LAST
    };
  
  /// Attribute flags; pass combination of these to ctor or setAttrs().
  enum {
    ANCHOR_PAGERMASK    = 0x0000FFFF	/// Bits reserved for pager attributes
   ,ANCHOR_UNDERLAY     = 0x80000000	/// Render this anchor under the text (otherwise on top)
   // The following flags indicate how to resolve ambiguity in display position when the
   // position lies at the split point of a split line.
   ,ANCHOR_S_ATEND      = 0x40000000	/// Starting point of this anchor is rendered at end
   					/// of row, else at start of next row.
   ,ANCHOR_E_ATEND      = 0x20000000	/// Ending point rendered at end else start.
   ,ANCHOR_REPEL_CURSOR = 0x10000000	/// Prohibit cursor from moving into this bracket.
   					/// This has the effect of not allowing the contents
   					/// to be altered by the user.
   ,ANCHOR_NO_DISAPPEAR = 0x08000000	/// Do not request self destruct from query() if set
   					/// to zero length.
    };
  
protected:
  FXTextAnchor();
private:
  FXTextAnchor(const FXTextAnchor&);
  FXTextAnchor& operator=(const FXTextAnchor&);
  
public:
  /// Construct with owner, attributes, user data, pos and epos.
  FXTextAnchor(FXTextSrc * ts, FXuint atr = 0, void * d = NULL, FXint sp = 0);
  
  /// Make an copy of this object, adding to same owner.
  virtual FXTextAnchor * clone();
  
  /// Query as to what to do when a replacement affects characters adjacent to or inside
  /// this anchor.  Return 1 to request deletion, 0 to keep.  adj_s, adj_e are suggested 
  /// adjustments to the start and end positions of this object, returned by this function.
  virtual FXint query(const FXTextReplacement * tr, FXint & adj_s, FXint & adj_e);
  
  /// Query as to whether this object is interested in the cursor position and/or any
  /// text replacement operations that might affect it.  If TRUE is returned, and it is
  /// the smallest enclosing bracket for an operation, then it given the opportunity by its 
  /// owner to have a say on cursor movement and text replacement.
  virtual FXbool isInterested(void) const;
  
  /// Query as to whether a cursor movement is permitted.  The answer should be TRUE if
  /// permitted.  If not null, pl is the widget which owns the cursor.  from and to
  /// are relative to the anchor start position.
  virtual FXbool permitsCursorMove(FXint from, FXint to, const FXPLText * pl = NULL) const;
  
  /// Query as to whether a text replacement is permitted.  The answer should be TRUE if
  /// permitted.  If not null, pl is the widget which is requesting the replacement.
  /// The parameters of tr are set up relative to the anchor start.
  virtual FXbool permitsReplacement(const FXTextReplacement * tr) const;
  
  /// Query as to whether this anchor is visible, disregarding the current window, in
  /// the nominated widget (or any widget if pl is null).
  virtual FXbool isVisible(const FXPLText * pl = NULL) const;
  
  /// Set visibility.
  virtual void setVisible(FXbool vis = TRUE);
  
  /// Return whether this object has a non-null glyph i.e. render() may actually draw
  /// something; either in the main text area, the aux area or both.
  virtual FXbool isRenderable() const;
  
  /// Return whether this object has a non-null glyph in the aux area.
  virtual FXbool isAuxRenderable() const;
  
  /// Render in a window.  'where' is 0 for main text display area, or 1 for auxiliary area.
  /// dc is a rendering context which has been locked onto the drawable.  x and y are the
  /// co-ordinates of the starting character baseline (left).  x2 and y2 are the ending
  /// char baseline (right).  flgs contains bits to indicate whether the start and/or
  /// end points are the real start and end of the bracket, or just the end of a row for
  /// multi-row brackets.  pl is the widget which is calling for the render.
  /// charwidth, -height and -base are the current character dimensions.
  virtual void render(FXint where, FXPLText * pl, FXDCWindow & dc, FXint x, FXint y, 
                      FXint x2, FXint y2, FXuint flgs,
                      FXint charwidth, FXint charheight, FXint charbase) const;
  
  /// Return bounding box.  'where' as for render().  The bounding box indicates the area
  /// that we render into, beyond the horizontal line defining the character baseline of
  /// the character(s) that we are bracketing.  For anchors, this degenerates to a point,
  /// so the BB should indicate the expansion of the point on the left side baseline of
  /// a character.  The returned x,y would usually both be negative to indicate the upper
  /// left corner of the BB.  pl is the widget which is calling for the render. 
  /// charwidth, -height and -base are the current character dimensions.
  virtual FXRectangle bbox(FXint where, FXPLText * pl,
                           FXint charwidth, FXint charheight, FXint charbase) const;

  /// Tell the specified page format about where we are, and update his attributes.
  /// start- and endpos indicate the range of char positions of interest. [start..end).
  /// The page format in turn notifies the widget.
  virtual void tellAttrs(FXPageFormat * pf, FXint startpos, FXint endpos) const;
  
  /// Update the current position of this anchor, then notify page formats which may be 
  /// tracking the location of this anchor.  If temp is true, graphical update deferred.
  virtual void move(FXint adj_s, FXint adj_e, FXbool temp);
  
  /// Return start position
  FXint getPos() const { return pos; }
  
  /// Return end position
  FXint getEpos() const { return epos; }
  
  /// Return number of characters covered
  FXint getLength() const { return epos - pos; }
  
  /// Return attributes
  FXuint getAttrs() const { return attrs; }
  
  /// Set attributes
  virtual void setAttrs(FXuint atr);
  
  /// Get mutexes
  virtual FXuint getMutex() const;
  
  /// Return FXTextSrc owner
  FXTextSrc * getTextSrc() const { return owner; }

  /// Get the current colour.  Always 'black' in base class.
  virtual FXColor getColor() const;
  
  /// Set the colour.  Page formats are notified in case their widgets are displaying this
  /// anchor. 
  virtual void setColor(FXColor c); 
  
  
  /// Save to a stream
  virtual void save(FXStream& store) const;

  /// Load from a stream
  virtual void load(FXStream& store);

  /// Destructor.
  virtual ~FXTextAnchor();
  };
  


/** FXTextCursor implements the cursor position and glyph for a text widget.  A particular
* text source may have more than one cursor (if it is being shared by several widgets),
* however cursor objects only display themselves in the text widget which created them,
* otherwise the user would be confused by the multiple cursors.
*/


class FXAPI FXTextCursor : public FXTextAnchor { 
  FXDECLARE(FXTextCursor)
private:
  FXPLText * pltext;	// Widget which created this cursor.
  FXuint style;		// Rendering style.
  FXshort hbeam;	// Half-size of top and bottom of I-beam (pixels horizontal)
  FXColor color;	// Rendering colours.
  FXColor lighterColor;	// Rendering colours.
  FXColor darkerColor;	// Rendering colours.
  FXuint flashon;	// Flash ON interval in milliseconds
  FXuint flashoff;	// Flash OFF interval in milliseconds
  FXuint on_style;	// Style when ON
  FXuint off_style;	// Style when OFF
  FXTimer * flashtimer;	// Cursor flash timer (NULL if steady).
  FXuint mutex;		// Visibility mutual exclusion.  If any anchor is setVisible(TRUE),
  			// then all other anchors with at least one matching mutex bit
  			// will be automatically setVisible(FALSE).  Flags taken from
  			// enum in base class.

public:
  /// Cursor rendering styles
  enum {
    TEXTCURSOR_IBEAM
   ,TEXTCURSOR_CARET
   ,TEXTCURSOR_BLOCK
   ,TEXTCURSOR_FRAME		// Hollow block
   ,TEXTCURSOR_TRIANGLE		// Hollow caret
   
   ,TEXTCURSOR_INVISIBLE = ~0
    };

  enum{
    ID_FLASHTIMER=FXTextAnchor::ID_LAST,
    ID_LAST
    };
  long onTimer(FXObject*,FXSelector,void*);
  
protected:
  FXTextCursor();
private:
  FXTextCursor(const FXTextCursor&);
  FXTextCursor& operator=(const FXTextCursor&);
  
public:
  /// Construct with owner (text source and widget), attributes, user data, pos and render style.
  FXTextCursor(FXTextSrc * ts, FXPLText * pl, FXuint atr = 0, void * d = NULL, 
               FXint sp = 0, FXuint rstyle = TEXTCURSOR_IBEAM);
  
  virtual FXbool isVisible(const FXPLText * pl = NULL) const;
  virtual void setVisible(FXbool vis = TRUE); 
  virtual FXbool isRenderable() const;
  virtual FXbool isAuxRenderable() const; 
  virtual void move(FXint adj_s, FXint adj_e, FXbool temp);
  virtual FXuint getMutex() const;
  
  /// Render according to current style (I-beam or caret).
  virtual void render(FXint where, FXPLText * pl, FXDCWindow & dc, FXint x, FXint y, 
                      FXint x2, FXint y2, FXuint flgs,
                      FXint charwidth, FXint charheight, FXint charbase) const;

  /// Return bounding box relative to anchor basepoint.
  virtual FXRectangle bbox(FXint where, FXPLText * pl,
                           FXint charwidth, FXint charheight, FXint charbase) const;

  /// Set attributes.  Only the widget's page format is notified.
  virtual void setAttrs(FXuint atr);
  
  /// Set the colour.  Only the widget's page format is notified. 
  virtual void setColor(FXColor c); 
  
  /// Get the current colour.  Returns the basic rendering colour.
  virtual FXColor getColor() const;
  
  /// Get the current style.
  FXuint getStyle() const { return on_style; }
  
  /// Set the style.  Simple, non-flashing.
  void setStyle(FXuint s);
  
  /// Set fancy style.  On/off times are in ms.  If either is zero, then equivalent to setStyle(onstyle).
  /// If both ontime and offtime are non-zero, then starts flashing at the specified rate.
  void setStyle(FXuint ontime, FXuint offtime, FXuint onstyle, FXuint offstyle = TEXTCURSOR_INVISIBLE);
  
  /// Start or stop flashing.  If setstyle, then also ensures cursor reset to normal (on) style.  If fancy
  /// ontime or offtime have not been specified (zero), then defaults on 750ms on/250ms off are set.
  void flash(FXbool on = TRUE, FXbool setstyle = TRUE);
  
  FXApp * getApp() const;
  
  /// Save to a stream
  virtual void save(FXStream& store) const;

  /// Load from a stream
  virtual void load(FXStream& store);

  /// Destructor.
  virtual ~FXTextCursor();
  };


/** Base class for text anchors that have a non-zero extent i.e. brackets.
 */

class FXAPI FXTextBracket : public FXTextAnchor { 
  FXDECLARE(FXTextBracket)
private:

public:
  enum {
    ID_LAST = FXTextAnchor::ID_LAST
    };
  
  
protected:
  FXTextBracket();
private:
  FXTextBracket(const FXTextBracket&);
  FXTextBracket& operator=(const FXTextBracket&);
  
public:
  /// Construct with owner, attributes, user data, start and end pos.
  FXTextBracket(FXTextSrc * ts, FXuint atr = 0, void * d = NULL, 
               FXint sp = 0, FXint ep = 0);  
  virtual FXTextAnchor * clone();

  virtual FXint query(const FXTextReplacement * tr, FXint & adj_s, FXint & adj_e);
  virtual void move(FXint adj_s, FXint adj_e, FXbool temp);
    
  /// Save to a stream
  virtual void save(FXStream& store) const;

  /// Load from a stream
  virtual void load(FXStream& store);
  };
  


/** Bracket used for text selections.  Encapsulates the selection mechanism.  Treated
 *  specially by text sources, since there must be just zero or one selection per
 *  text source.  All members are protected (except to our FXTextSrc friend), since
 *  access should be performed indirectly via the text source.
 */

class FXAPI FXTextSelection : protected FXTextBracket { 
  FXDECLARE(FXTextSelection)
  
protected:
  FXTextSelection();
private:
  FXTextSelection(const FXTextSelection&);
  FXTextSelection& operator=(const FXTextSelection&);
  
protected:

  friend class FXTextSrc;
  
  /// Construct with owner, attributes, user data, start and end pos.
  FXTextSelection(FXTextSrc * ts, FXuint atr = 0, void * d = NULL, 
               FXint sp = 0, FXint ep = 0);  
  
  };
  

/** Bracket which is designed to accept user input e.g. in a terminal mode
*   application.
*
*   In addition, SEL_IO_WRITE messages are handled by sending the bracket
*   contents to the socket, then self-resetting to zero size at the end
*   of text (taking the cursor with it).  FIXME - other behaviours useful.
*/

class FXAPI FXInputBracket : public FXTextBracket { 
  FXDECLARE(FXInputBracket)
public:
  enum {
    ID_LAST
    };
  // SEL_IO_WRITE/ID_SOCKET message from an FXInputHandle.
  long onIOWrite(FXObject*,FXSelector,void*);
  
protected:
  FXInputBracket();
private:
  FXInputBracket(const FXInputBracket&);
  FXInputBracket& operator=(const FXInputBracket&);
  FXint streamno;		// Process stream number (-1 if none).
  
public:

  friend class FXTextSrc;
  
  /// Construct with owner, socket to write to, attributes, user data, start and end pos.
  FXInputBracket(FXTextSrc * ts, FXint stream, FXuint atr = 0, void * d = NULL, 
               FXint sp = 0, FXint ep = 0);  
  
  virtual FXint query(const FXTextReplacement * tr, FXint & adj_s, FXint & adj_e);
  virtual FXbool isInterested(void) const;
  virtual FXbool permitsReplacement(const FXTextReplacement * tr) const;
  };
  


/** Bracket which is designed to accept application output but not allow the
*   user to overwrite it.  It does this by prohibiting the cursor from being
*   placed within its bounds.
*/

class FXAPI FXOutputBracket : public FXTextBracket { 
  FXDECLARE(FXOutputBracket)
public:
  enum {
    ID_LAST
    };
  
protected:
  FXOutputBracket();
private:
  FXOutputBracket(const FXOutputBracket&);
  FXOutputBracket& operator=(const FXOutputBracket&);
  
public:

  friend class FXTextSrc;
  
  /// Construct with owner, attributes, user data, start and end pos.
  FXOutputBracket(FXTextSrc * ts, FXuint atr = 0, void * d = NULL, 
               FXint sp = 0, FXint ep = 0);  
  virtual FXTextAnchor * clone();
  
  virtual FXbool isInterested(void) const;
  virtual FXbool permitsCursorMove(FXint from, FXint to, const FXPLText * pl = NULL) const;
  virtual FXbool permitsReplacement(const FXTextReplacement * tr) const;
  };
  


#endif