/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                              Text Buffer                                      *
*********************************************************************************
*   $Id: FXTextBuf.cpp,v 1.3 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXTextBuf.cpp,v $
* Revision 1.3  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.2  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.1.1.1  2000/10/29 14:20:36  tekno
* Initial release
*
*
*
********************************************************************************/
#ifdef __INSURE__
#define __attribute__(x)
#endif

#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"

#include "FXTextMap.h"
#include "FXTextBuf.h"

FXTextBuf::~FXTextBuf() { 
  }


/*--------------- String-based buffer with gap ------------------*/

/* NB: we add a 'virtual' newline at the end of the string, which is not actually stored
   but appears to be there if the API is used.  This is a requirement for using the
   FXTextMap object to split the string into lines, since each line _must_ be terminated
   with a newline; in particular the last line.
 */

#define MAX_GAP   4096	// Max allowed gap size before we reallocate smaller
#define START_GAP 1024	// Initial gap size
#define INC_GAP   1024	// Extra gap size when allocating more

FXTextBufS::FXTextBufS(const FXString & init) {
  ok = TRUE;
  slen = init.length();
  size = slen + START_GAP;
  gs = slen;	// Gap at end
  FXMALLOC(&s, FXchar, size);
  if (!s)
    ok = FALSE;
  else
    memcpy(s, init.text(), slen);
  }


FXTextBufS::FXTextBufS(const FXchar * init, FXuint len) {
  ok = TRUE;
  slen = len;
  size = slen + START_GAP;
  gs = slen;	// Gap at end
  FXMALLOC(&s, FXchar, size);
  if (!s)
    ok = FALSE;
  else
    memcpy(s, init, slen);
  }


FXchar FXTextBufS::get(FXuint offset) const {
  if (offset == slen)
    return '\n';
  if (offset > slen)
    return 0;
  return offset < gs ? s[offset] : s[offset+size-slen];
  }


FXuint FXTextBufS::find(FXuint offset, FXchar delim /*= '\n'*/ ) const {
  FXuint i;
  FXuint n = 1;
  if (offset < gs) {
    for (i = offset; i < gs; i++, n++)
      if (s[i] == delim)
        return n;
    for (i = gs+size-slen; i < size; i++, n++)
      if (s[i] == delim)
        return n;
  }
  else
    for (i = offset+size-slen; i < size; i++, n++)
      if (s[i] == delim)
        return n;
  return slen - offset + 1;
  }
  
void FXTextBufS::condense() {
  if (gs < slen) {
    memmove(s+gs, s+gs+size-slen, slen-gs);
    gs = slen;
    }
  }

void FXTextBufS::moveGap(FXuint g) {
  if (g >= slen) condense();
  else {
    FXint del = (FXint)g - (FXint)gs;
    if (del > 0) {
      memmove(s+gs, s+gs+size-slen, del);
      gs += del;
      }
    else {
      memmove(s+gs+size-slen+del, s+gs+del, -del);
      gs += del;
      }
    }
  }

FXuint FXTextBufS::getmem(FXuint offset, FXuint len, FXchar * mem) const {
  ((FXTextBufS *)this)->condense();	// Discard const
  if (offset+len <= slen) {
    memcpy(mem, s+offset, len);
    return len;
    }
  else if (offset+len == slen+1) {
    memcpy(mem, s+offset, len-1);
    mem[len-1] = '\n';
    return len;
    }
  else if (offset < slen+1) {
    memcpy(mem, s+offset, slen-offset);
    mem[slen-offset] = '\n';
    return slen+1-offset;
    }
  return 0;
  }


FXuint FXTextBufS::length() const {
  return slen + 1;	// Add 1 for virtual newline
  }


FXuint FXTextBufS::alloc() const {
  return size;
  }


FXuint FXTextBufS::getContig(FXuint offset, FXchar * & ptr) const {
  if (offset > slen) {
    ptr = 0;
    return 0;
    }
  else if (offset == slen) {
    ptr = "\n";	// Point to a real virtual newline :-)
    return 1;
    }
  else if (offset >= gs) {
    ptr = s + offset + size-slen;
    return slen - offset;
    }
  ptr = s + offset;
  return gs - offset;
  }


FXuint FXTextBufS::getBackwardsContig(FXuint & offset, FXchar * & ptr) const {
  if (offset > slen+1 || !offset) {
    ptr = 0;
    return 0;
    }
  else if (offset == slen+1) {
    ptr = "\n";	// Point to a real back-to-front virtual newline :-)
    offset--;
    return 1;
    }
  else if (offset > gs) {
    ptr = s+gs+size-slen;
    offset = gs;
    return slen-gs;
    }
  ptr = s;
  offset = 0;
  return gs;
  }


void FXTextBufS::replace(FXuint offset, FXuint len, const FXchar * what, FXuint rlen) {
  // Note that the virtual newline cannot be replaced!
  FXASSERT(offset <= slen);
  FXASSERT(offset + len <= slen); 
  if (offset > slen)
    offset = slen, len = 0;
  else if (offset + len > slen)
    len = slen - offset;
  
  moveGap(offset);
  slen -= len;		// Effectively delete old section
  
  if (slen+rlen > size) {
    // Need to reallocate larger
    FXuint newsize = slen+rlen+INC_GAP;
    FXuint atend = slen - gs;
    FXRESIZE(&s, FXchar, newsize);
    if (!s) {
      ok = FALSE;
      return;
      }
    memmove(s+newsize-atend, s+size-atend, atend);
    size = newsize;
    }
    
  memcpy(s+gs, what, rlen);
  gs += rlen;		// Effectively insert new section 
  slen += rlen;
  
  if (size-slen > MAX_GAP) {
    // Gap too big, resize downwards
    FXuint newsize = slen+INC_GAP;
    FXuint atend = slen - gs;
    memmove(s+newsize-atend, s+size-atend, atend);
    FXRESIZE(&s, FXchar, newsize);
    if (!s) {
      ok = FALSE;
      return;
      }
    size = newsize;
    }
  }


FXTextBufS::~FXTextBufS() {
  FXFREE(&s);
  }




/*----------- Paging-based non-linear buffer -----------*/
//FIXME: None of this is tested!  Needs more code for error handling.

const FXuint FXTextBufP::pagesize = FXPAGESIZE;

FXuint FXTextBufP::allocPage(FXuint startpage, FXuint numpage, FXuint req) {
  FXuint i;
  
  // Use malloc, so get exact amount requested
  for (i = startpage; i < startpage+numpage; i++) {
    if (!FXMALLOC(&pagetab[i].base, FXchar, req))
      return 0;
    pagetab[i].gap = 0;
    map->updateLeaf(i<<1, 0, req);
    }
  return req;
  }


FXTextBufP::FXTextBufP(const FXString & init) {
  FXuint len = init.length();
  FXuint i, inbuf, got;
  
  pages = (len + pagesize - 1)/pagesize;
  FXCALLOC(&pagetab, FXPageDesc, pages);
  FXASSERT(pagetab);
  map = new FXTextMap(pages);
  FXASSERT(map);
  if (!(got = allocPage(0, pages, pagesize)))
    FXASSERT(0);
  for (i = 0; i < pages; i++) {
    memcpy(pagetab[i].base, init.text() + i*pagesize, inbuf = FXMIN(pagesize, len - i*pagesize));
    pagetab[i].gap = inbuf;
    map->updateLeaf(i<<1, inbuf, got);
    }
  }


FXTextBufP::FXTextBufP(const FXchar * init, FXuint len) {
  FXuint i, inbuf, got;
  
  pages = (len + pagesize - 1)/pagesize;
  FXCALLOC(&pagetab, FXPageDesc, pages);
  FXASSERT(pagetab);
  map = new FXTextMap(pages);
  FXASSERT(map);
  if (!(got = allocPage(0, pages, pagesize)))
    FXASSERT(0);
  for (i = 0; i < pages; i++) {
    memcpy(pagetab[i].base, init + i*pagesize, inbuf = FXMIN(pagesize, len - i*pagesize));
    pagetab[i].gap = inbuf;
    map->updateLeaf(i<<1, inbuf, got);
    }
  }


FXchar FXTextBufP::get(FXuint offset) const {
  if (offset >= map->root_Pn())
    return 0;
  FXTextNode search(offset, 0);
  FXuint node = map->modulo_Pn(search);
  FXuint pgent = node >> 1;		// page table entry index
  FXuint nchars = map->node_Pn(node);	// # chars in this entry
  if (search.Pn >= pagetab[pgent].gap)
    return pagetab[pgent].base[search.Pn + map->node_Ln(node) - nchars];
  return pagetab[pgent].base[search.Pn];
  }


FXuint FXTextBufP::find(FXuint offset, FXchar delim /*= '\n'*/) const {
  FXuint len = 0;
  FXchar * ptr, * ptr2;
  FXuint clen;
  
  while ((clen = getContig(offset + len, ptr))) {
    if ((ptr2 = (FXchar *)memchr(ptr, delim, clen)))
      return (ptr2 - ptr) + 1 + len;
    len += clen;
    }
  return len;
  }


FXuint FXTextBufP::getmem(FXuint offset, FXuint dlen, FXchar * mem) const {
  FXuint len = 0;
  FXchar * ptr;
  FXuint clen;
  
  while (len < dlen && (clen = getContig(offset + len, ptr))) {
    if (clen > dlen - len) clen = dlen - len;
    memcpy(mem + len, ptr, clen);
    len += clen;
    }
  return len;
  }


FXuint FXTextBufP::length() const {
  return map->root_Pn();
  }


FXuint FXTextBufP::alloc() const {
  return map->root_Ln();
  }


FXuint FXTextBufP::getContig(FXuint offset, FXchar * & ptr) const {
  if (offset >= map->root_Pn()) {
    ptr = 0;
    return 0;
    }
  FXTextNode search(offset, 0);
  FXuint node = map->modulo_Pn(search);
  FXuint pgent = node >> 1;		// page table entry index
  FXuint nchars = map->node_Pn(node);	// # chars in this entry
  FXuint nalloc = map->node_Ln(node);	// allocated chars in this entry
  if (search.Pn >= pagetab[pgent].gap) {
    ptr = pagetab[pgent].base + (search.Pn + nalloc - nchars);
    return nchars - search.Pn;
    }
  ptr = pagetab[pgent].base + search.Pn;
  return pagetab[pgent].gap - search.Pn;
  }

FXuint FXTextBufP::getBackwardsContig(FXuint & offset, FXchar * & ptr) const {
  return 0;	//FIXME 
  }

void FXTextBufP::moveGap(FXuint pgent, FXuint offset) {
  FXuint g = pagetab[pgent].gap;
  FXchar * b = pagetab[pgent].base;
  FXchar gs = map->node_Ln(pgent<<1) - map->node_Pn(pgent<<1);
  if (offset < g)
    memmove(b+offset+gs, b+offset, g-offset);
  else if (offset > g)
    memmove(b+g, b+g+gs, offset-g);
  pagetab[pgent].gap = offset;
  }

FXuint FXTextBufP::spaceAvail(FXint dir, FXuint start, FXuint req) {
  FXint node = (FXint)(start<<1);
  FXint last = dir > 0 ? map->highLeaf() : 0;
  FXint incr = dir*2;
  
  while (node != last) {
    node += incr;
    FXuint nalloc = map->node_Ln(node);
    if (nalloc - map->node_Pn(node) >= req)
      return (FXuint)node >> 1;
    req -= nalloc - map->node_Pn(node);
    }
  return start;
  }

void FXTextBufP::insertPages(FXuint fromPage, FXuint numpages, FXuint req) {
  // Insert into pagetab.  Reallocate if necessary.
  if (map->getN() + numpages > pages) {
    FXuint newsize = map->getN() + numpages + 128;	// Allow extra growth of 128
    if (!FXRESIZE(&pagetab, FXPageDesc, newsize))
      FXASSERT(0);
    pages = newsize;
    }
  memmove(pagetab+fromPage+numpages, pagetab+fromPage,
            sizeof(*pagetab) * (map->getN() - fromPage));
  memset(pagetab+fromPage, 0, sizeof(*pagetab) * numpages);
  if (!map->insertLeaves(fromPage<<1, numpages, NULL))
    FXASSERT(0);
  
  if (!allocPage(fromPage, numpages, req))
    FXASSERT(0); 
    
  }


void FXTextBufP::replace(FXuint offset, FXuint len, const FXchar * what, FXuint rlen) {
  FXuint clen;
  FXchar * ptr;
  FXuint rl, d;

  if (offset > map->root_Pn())
    offset = map->root_Pn(), len = 0;
  else if (offset + len > map->root_Pn())
    len = map->root_Pn() - offset;
  
  /* Do the replacement-only part */
  FXuint repllen = FXMIN(len, rlen);
  while (repllen && (clen = getContig(offset, ptr))) {
      memcpy(ptr, what, rl = FXMIN(repllen, clen));
      repllen -= rl;
      what += rl;
      offset += rl;
      rlen -= rl;
      len -= rl;
      }
  
  /* If exact replacement, our work is done */
  if (!rlen && !len)
    return;
  
  /* Now the replacement part is done (only one of rlen or len is non-zero), 
     work out the nett insertion */
  FXbool insert = rlen > len;
  if (insert)
    d = rlen;
  else
    d = len;
  
  FXTextNode search(offset, 0);
  FXuint node = map->modulo_Pn(search);
  FXuint pgent = node >> 1;		// page table entry index
  FXuint nchars = map->node_Pn(node);	// # chars in this entry
  FXuint nalloc = map->node_Ln(node);	// allocated chars in this entry
  
  if (insert) {
    // move gap to the insertion point.  We call the data on this page before the gap B,
    // and the data after the gap A.
    moveGap(pgent, search.Pn);
    if (d <= nalloc - nchars) {
      // Can fit in this page's gap
fit_in_gap:
      memcpy(pagetab[pgent].base + search.Pn, what, d);
      pagetab[pgent].gap += d;
      map->updateLeaf(node, nchars + d, nalloc);
      }
    else {
      /* Can't fit in gap.  Move as much of the data to the right of the IP, A, into the next
         page as possible.  If still can't fit the insertion, logically tack the remaining A
         to the end of the insert data.  Move as much data as possible from the
         start of the insert data into the new (possibly bigger) gap.  The remaining insert
         data is put into new pages inserted after the current page.
       */
      FXuint a = nchars - search.Pn;	// # chars after insertion point (A)
      if (a && node < map->highLeaf()) {
        FXuint next_nchars = map->node_Pn(node+2);
        FXuint next_nalloc = map->node_Ln(node+2);
        FXuint next_space = next_nalloc - next_nchars;
        if (next_space) {
          // move some or all of A into next page and update map
          FXuint chars2move = FXMIN(a, next_space);
          moveGap(pgent+1, 0);	// Put gap at start
          FXchar * b = pagetab[pgent+1].base + next_nalloc - next_nchars - chars2move;
          FXchar * c = pagetab[pgent].base;
          memcpy(b, c + nalloc - chars2move, chars2move);
          FXuint rem = a - chars2move;
          if (rem)
            memmove(c + nalloc - rem, c + nalloc - a, rem);
          map->updateLeaf(node+2, next_nchars + chars2move, next_nalloc);
          nchars -= chars2move;
          map->updateLeaf(node, nchars, nalloc);
          a -= chars2move;
          }
        // Now a is remaining chars on current page after insertion point.
        if (d <= nalloc - nchars)
          goto fit_in_gap;	// we created enough space.
        
        // Try moving B to the previous page (not implemented)
        /*...*/
        
        FXuint da = d + a - (nalloc - search.Pn);	// total amount to fit in new pages
        FXASSERT(da > 0);
        FXuint newpages = roundToPage(da);
        // Create new pages and update map and pagetab.  Pages inserted after current page.
        // Gaps are at start of page.
        insertPages(pgent+1, newpages, pagesize);
        // Put the remainder, A, at the end of the last page.
        if (a) {
          FXchar * b = pagetab[pgent+newpages].base;
          FXuint last_nalloc = map->node_Ln(pgent+newpages << 1);
          memcpy(b + last_nalloc - a, pagetab[pgent].base + nalloc - a, a);
          map->updateLeaf(pgent+newpages << 1, a, last_nalloc);
          }
        // Put insert data into current page, filling it.
        memcpy(pagetab[pgent].base + search.Pn, what, nalloc - search.Pn);
        what += nalloc - search.Pn;
        d -= nalloc - search.Pn;
        map->updateLeaf(pgent<<1, nalloc, nalloc);	// Current page now full
        // Put rest of data into following pages.
        FXuint pg = pgent;
        while (d) {
          pg++;
          FXuint nchars = map->node_Pn(pg<<1);
          FXuint nalloc = map->node_Ln(pg<<1);
          FXuint chars2move = FXMIN(nalloc, d-nchars);
          memcpy(pagetab[pg].base, what, chars2move);
          pagetab[pg].gap += chars2move;
          map->updateLeaf(pg<<1, chars2move + nchars, nalloc);
          }
        FXASSERT(pg == pgent + newpages);
        }
      }
    }
  else {
    // Nett deletion 
    }
  
  }



FXTextBufP::~FXTextBufP() {
  if (map) delete map;
  map = (FXTextMap *)-1;
  if (pagetab) FXFREE(&pagetab);
  pagetab = (FXPageDesc *)-1;
  }

