/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                              Text Buffer                                      *
*********************************************************************************
*   $Id: FXTextBuf.h,v 1.2 2000/11/01 11:27:10 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXTextBuf.h,v $
* Revision 1.2  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.1.1.1  2000/10/29 14:19:46  tekno
* Initial release
*
*
*
********************************************************************************/
#ifndef FXTEXTBUF_H
#define FXTEXTBUF_H

#include "FXString.h"

/* For OSs (such as Linux) which support paging I/O and mmap(), we want to be
* cogniscent of the page size so that we can make more efficient use of the
* underlying virtual memory mechanism.  On other OSs, we just use a page size
* of 4096.
*/
#ifdef EXEC_PAGESIZE
#define FXPAGESIZE EXEC_PAGESIZE
#else
#define FXPAGESIZE 4096
#endif

class FXTextMap;

/** Base class for FXTextSrc text buffering strategy.
* The base class itself only defines the interface and is not instantiable.  
* Derived class FXTextBufP (P for Page) implements a paged buffer suitable for 
* arbitrarily large text sources.
* FXTextBufS (S for String) implements a memory buffer, which would be suitable
* for small text sources, or when update efficiency is not significant.  This memory
* buffer implements a single gap, as done by the FXText class.
* These classes are not usually included by user programs; they are intended for
* implementors of FXTextSrc and FXPageFormat classes.
*/

class FXTextBuf {
protected:
  FXbool ok;		// Flag to indicate construction completed 
  			// successfully, and object state is usable.
  
public:
  
  /// Test whether constructor completed successfully.  Since ctors may allocate memory etc.,
  /// it is necessary for the caller to check whether construction was sucessful.  If not,
  /// then the dtor should be invoked as soon as possible and an error flagged to the user.
  /// Note that the ctor is exclusively responsible for setting this flag.  The flag may
  /// be reset by member functions which encounter errors.
  FXbool isOk() const { return ok; }
  
  /// Get character at given offset.  If offset is outside buffer, return zero.
  virtual FXchar get(FXuint offset) const = 0;
  
  /// Return number of bytes required to contain chars starting at offset, up to and
  /// including the next occurrence of delim.  If delim is not found before end of
  /// buffer, return the number of chars from offset to end of buffer.
  virtual FXuint find(FXuint offset, FXchar delim = '\n') const = 0;
  
  /// Retrieve len bytes starting at offset, into mem.  Returns number of bytes moved,
  /// which may be less than len if end of buffer encountered.
  virtual FXuint getmem(FXuint offset, FXuint len, FXchar * mem) const = 0;
  
  /// Return number of bytes currently in buffer
  virtual FXuint length() const = 0;
  
  /// Return allocated space in buffer, which may be greater than length().
  virtual FXuint alloc() const = 0;
  
  /// Obtains a pointer to the actual buffer storage of the byte at offset.  The number
  /// of bytes which may be retrieved using e.g. memcpy() is returned.  The pointer is
  /// only guaranteed to be valid until the next text buffer operation.  If there is no
  /// data at the specified offset, then the pointer is set to null and zero returned.
  /// Otherwise, the returned length is guaranteed to be non-zero.
  virtual FXuint getContig(FXuint offset, FXchar * & ptr) const = 0;
  
  /// Similar to getContig() except moves towards the start of the buffer.  On return,
  /// ptr points to the first byte addressable of a contiguous stretch which is guaranteed
  /// to extend to at least offset-1.  Returns the actual number of contiguously accessible
  /// bytes, which may extend to offset or beyond.  If offset is past the (end of buffer)+1,
  /// or at the start (0), then 0 is returned.  Otherwise, the returned length is guaranteed 
  /// to be non-zero.  Unlike getContig, the offset parameter is updated to the offset of the
  /// start of the return data.
  virtual FXuint getBackwardsContig(FXuint & offset, FXchar * & ptr) const = 0;
  
  /// Replace a range of bytes starting at offset of length len in the buffer.  what points
  /// to the replacement data and rlen is the length of that data.  If len is zero, then
  /// the replacement data is inserted immediately before the byte at offset.  If offset
  /// is greater than length(), then it is deemed to be length() and len is deemed zero.
  /// If len exceeds the remaining buffer length, only the trailing part of the buffer
  /// from offset is replaced.
  virtual void replace(FXuint offset, FXuint len, const FXchar * what, FXuint rlen) = 0;
  
  /// Append data.
  void append(const FXchar * what, FXuint rlen) { replace(length()+1, 0, what, rlen); }
  
  /// Prepend data.
  void prepend(const FXchar * what, FXuint rlen) { replace(0, 0, what, rlen); }
  
  /// Destructor.
  virtual ~FXTextBuf();
  };
  

/** Implementation of FXTextBuf which uses a single buffer with a 'gap'.  There are thus
    two areas:  Area 1 starts at 's' and has a length of 'gs'.  The gap is then size-slen.
    Area 2 starts at s+gs+size-slen, and has a length of size-(gs+size-slen) = slen-gs.
    Thus the areas are at the start and end of the allocated buffer, and a gap of zero or
    more locations exists between them.  The location and size of the gap are automatically
    managed by this class, depending on the update operations applied.  The required 'virtual
    newline' is not actually stored, but forms a logical third extent of the buffer.
*/
class FXTextBufS : public FXTextBuf {
  FXchar * s;	// The buffer area
  FXuint slen;	// Cached length of string (we need it often)
  FXuint size;	// Allocated size
  FXuint gs;	// Offset of start of gap
  
  void condense(); // Squeeze out gap
  void moveGap(FXuint g); // Move gap to offset g

public:
  /// Construct by copying an existing FXString.
  FXTextBufS(const FXString & init);
  
  /// Construct by copying from memory.
  FXTextBufS(const FXchar * init, FXuint len);
  
  FXchar get(FXuint offset) const;
  FXuint find(FXuint offset, FXchar delim = '\n') const;
  FXuint getmem(FXuint offset, FXuint len, FXchar * mem) const;
  FXuint length() const;
  FXuint alloc() const;
  FXuint getContig(FXuint offset, FXchar * & ptr) const;
  FXuint getBackwardsContig(FXuint & offset, FXchar * & ptr) const;
  void replace(FXuint offset, FXuint len, const FXchar * what, FXuint rlen);
  ~FXTextBufS();
  
  };
  
  

/// Implementation of FXTextBuf which uses a paging scheme and stores the data non-
/// contiguously.  When the OS supports it, it will make use of memory mapping.
class FXTextBufP : public FXTextBuf {
  FXTextMap * map;	// This map allows us to convert between byte offsets and
  			//  page numbers.  Pn elements will be byte counts, Ln will
  			//  be byte allocation counts (such that Ln >= Pn).
  static const FXuint pagesize;		// Must be power of 2.
  
  // Each of these structs represents a single page.  Note that the number of bytes stored
  // in the page is derived from the above map.  The total allocated length of the page
  // is stored in the Ln statistic of the map node, and may or may not equal pagesize
  // depending on the allocation strategy.
  struct FXPageDesc {
    FXchar * base;	// Page base address
    FXuint   gap;	// Offset of gap within page.  Bytes before the gap offset are
    			//  stored at the start of the page, those after are stored at
    			//  the end.  Maintaining a gap makes successive insertions more
    			//  efficient.
    };
  
  FXuint pages;		// Allocated number of entries in following array
  FXPageDesc * pagetab;	// Page address table.  There will be one entry for each leaf node
  			//  in the map i.e. dimensioned same as map->getN() <= pages.

  // Round up to next multiple of page size
  FXuint roundToPage(FXuint size) const { return (size-1 | pagesize-1) + 1; }
  
  // Allocate buffer page for given successive entry(s) in pagetab.  Returns number of
  // bytes allocated (per page) for success, or zero for failure.  req is the suggested
  // number of bytes per page, however more than this may be returned.  If req is a
  // multiple of pagesize, then exactly the requested amount will be returned.  Also
  // updates map with Pn=0 and Ln=allocated amount.
  FXuint allocPage(FXuint startpage, FXuint numpage, FXuint req);

  // Insert numpages new pages, each of size req, before the page beforePage.  This
  // maintains both map and pagetab.  The page gap is set to zero.
  // FIXME: need exception handling.
  void insertPages(FXuint fromPage, FXuint numpages, FXuint req);

  // Move the gap in given page to new offset.
  void moveGap(FXuint pgent, FXuint offset);
  
  // Return page entry of a page in direction dir (1=up, -1=down) from start which,
  // along with intermediate pages, collectively contain enough gap for req bytes
  // to be inserted.  The space in start itself is not counted.  If there is not enough
  // space in the specified direction, will return 'start'.
  FXuint spaceAvail(FXint dir, FXuint start, FXuint req);

public:
  /// Construct by copying an existing FXString.
  FXTextBufP(const FXString & init);
  
  /// Construct by copying from memory.
  FXTextBufP(const FXchar * init, FXuint len);
  
  FXchar get(FXuint offset) const;
  FXuint find(FXuint offset, FXchar delim = '\n') const;
  FXuint getmem(FXuint offset, FXuint len, FXchar * mem) const;
  FXuint length() const;
  FXuint alloc() const;
  FXuint getContig(FXuint offset, FXchar * & ptr) const;
  FXuint getBackwardsContig(FXuint & offset, FXchar * & ptr) const;
  void replace(FXuint offset, FXuint len, const FXchar * what, FXuint rlen);
  ~FXTextBufP();
  
  };
  
  


#endif