/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                           Text metric mapping                                 *
*********************************************************************************
*   $Id: FXTextMap.h,v 1.2 2000/11/01 11:27:10 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXTextMap.h,v $
* Revision 1.2  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.1.1.1  2000/10/29 14:19:22  tekno
* Initial release
*
*
*
********************************************************************************/
#ifndef FXTEXTMAP_H
#define FXTEXTMAP_H


/** FXTextNode struct is used as the binary tree nodes for mapping line numbers to
* character offsets in the FXTextBuf buffer.  Each member is the sum of the corresponding
* entries of its two children, except for leaf nodes where the value reflects the actual 
* buffer entry.
*/
struct FXTextNode {
  FXuint   Pn;		// Length of line(s) or rows(s) in child nodes or buffer
  FXuint   Ln;		// Number of lines/rows.   These lines are stored contiguously
  			//  and their total length is Pn.
  
  FXTextNode(FXuint p, FXuint l): Pn(p), Ln(l) {}
  };

/** Base class for FXTextSrc text metric mapping.  This implements a binary tree
* of text metrics (mapping between line/row numbers and character offsets) so that
* mapping operations are O(log(n)).  This class is not usually included by user
* programs; it is intended for implementors of FXTextSrc and FXPageFormat classes.
*/

class FXTextMap {

  FXbool ok;		// Flag to indicate construction completed successfully
  FXTextNode * tree;	// Point to first tree node (in linear array)
  FXuint   treealloc;	// Allocated entries in tree - is always 2**V, even though up to half
  			//  of the high entries may be zero.  This simplifies bookkeeping.
  FXuint   N;		// Number of non-zero leaf nodes (1st is 0 index, last is 2(N-1) index).
  FXuint   V;		// Number of levels (=ceil(log2(N))+1).  Level 0 is leaf nodes,
  			//  level 1 is parents of these etc.  Root node is at level V-1.
  FXuint   root;	// Index of root node (=2**(V-1)-1)
  FXint    shift;	// Amount of last shift (see below), 0 if no shift, -ve for delete,
			//  +ve for insert.
  FXuint   updnode;	// Index of first leaf node shifted (if shift != 0), or index of
			//  leaf node which was updated. 
  FXbool   dirty;	// UpdateNode() called, but rest of tree not updated yet.
  FXuint   lowupd;	// Lowest leaf node that was updated (if dirty)
  FXuint   highupd;	// Highest leaf node updated (if dirty)
  
protected:
  FXTextMap();
private:
  FXTextMap(const FXTextMap&);
  FXTextMap& operator=(const FXTextMap&);
  
public:
  /// Construct with initial size for tree.
  FXTextMap(FXuint leaves);

  /// Test whether constructor completed successfully.  Since ctors may allocate memory etc.,
  /// it is necessary for the caller to check whether construction was sucessful.  If not,
  /// then the dtor should be invoked as soon as possible and an error flagged to the user.
  /// Note that the ctor is exclusively responsible for setting this flag.
  FXbool isOk() const { return ok; }
  
protected:
  /*--- Macros ---*/
  /* These do not do any validation! */
  
  /// Find parent, given child at level
  FXuint parent(FXuint child, FXuint level) const { return child & ~(1<<level+1) | 1<<level; }
  
  /// Find lower (left) child, given parent at level
  FXuint lowchild(FXuint parent, FXuint level) const { return parent & ~(1<<level-1); }

  /// Find higher (right) child, given parent at level
  FXuint highchild(FXuint parent, FXuint level) const { return parent & ~(1<<level-1) | 1<<level;}
  
  /// Find lower peer at same given level
  FXuint lowpeer(FXuint node, FXuint level = 0) const { return node - (1<<level+1); }

  /// Find higher peer at same given level
  FXuint highpeer(FXuint node, FXuint level = 0) const { return node + (1<<level+1); }

  /*--- end of macros --- */

  /// Traverse tree upwards, updating ancestor nodes given that the specified inclusive range of
  /// leaf nodes has been updated.
  void updateRange(FXuint lowleaf, FXuint highleaf);
  
public:
  /// Traverse tree from root down searching for node corresponding to initial values.
  /// On call, search is set up with Ln field set to the desired line number.  On return,
  /// the Ln field will contain the remainder (non-zero only if line is not first in this
  /// node).  Pn will contain the character offset of the start of the _first_ line in this
  /// node with respect to the text source.  Returns the leaf node number in which found.
  FXuint modulo_Ln(FXTextNode & search);
  
  /// Traverse tree from root down searching for node corresponding to initial values.
  /// On call, search is set up with Pn field set to the desired char offset.  On return,
  /// the Pn field will contain the remainder (non-zero only if char is not first in this
  /// node).  Ln will contain the line number of the _first_ line in this node.  Returns 
  /// the leaf node number in which found.
  FXuint modulo_Pn(FXTextNode & search);
  
  /// Return Ln of root node.
  FXuint root_Ln();
  
  /// Return Pn of root node.
  FXuint root_Pn();
  
  /// Return Ln of given node.
  FXuint node_Ln(FXuint node);
  
  /// Return Pn of given node.
  FXuint node_Pn(FXuint node);
  
  /// Return highest numbered leaf node
  FXuint highLeaf() const { return (N<<1)-2; }
  
  /// Return number of leaf nodes
  FXuint getN() const { return N; }
  
  /// Insert new leaf node(s).  The new leaves are inserted before beforeLeaf,
  /// and num are inserted.  The new leaves are zerod if newLeaves is NULL, or the given
  /// array is used to set the statistics for the new leaves.  The rest of the
  /// tree is updated to be consistent.  Note that NewLeaves has its nodes contiguous,
  /// not only in even numbered entries as thay are stored in the tree.
  /// Returns FALSE iff failed.
  FXbool insertLeaves(FXuint beforeLeaf, FXuint num = 1, const FXTextNode * newLeaves = 0);
  
  /// Delete leaf node(s) from and including fromLeaf.  num leaves are deleted and the tree
  /// is updated to be consistent.
  void deleteLeaves(FXuint fromLeaf, FXuint num = 1);
  
  /// Change the statistics of the given leaf node.  This does not immediately update the
  /// tree, but marks it as needing update so that next call which requires consistent
  /// tree will automagically update everything.
  void updateLeaf(FXuint leaf, FXuint Pn, FXuint Ln);
  
  /// Destructor.
  ~FXTextMap();
  };
  



#endif