/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                              Text Source                                      *
*********************************************************************************
*   $Id: FXTextSrc.h,v 1.3 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXTextSrc.h,v $
* Revision 1.3  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.2  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.1.1.1  2000/10/29 14:18:27  tekno
* Initial release
*
*
*
********************************************************************************/
#ifndef FXTEXTSRC_H
#define FXTEXTSRC_H

#ifndef FXOBJECTLIST_H
#include "FXObjectList.h"
#endif
#ifndef FXARRAY_H
#include "FXElement.h"
#include "FXArray.h"
#endif

class FXApp;
class FXWindow;
class FXPLText;
class FXPageFormat;
class FXTextAnchor;
class FXOutputBracket;
class FXInputBracket;
class FXTextSelection;
class FXProcess;
struct FXFindData;

/** Base class for FXPLText widget data source.  Robust application code should call
* the isOK() member after calling any non-const member function (including ctors).  In
* particular, replace(), rollback(), rollforward() do not otherwise return error
* indicators.  If isOK() returns false, then the object state must be considered
* unstable and the destructor should be called.
*/

class FXAPI FXTextSrc : public FXObject {
  FXDECLARE(FXTextSrc)

protected:
  FXbool ok;				// Flag to indicate construction completed successfully
  FXObjectListOf<FXPageFormat> pfs;	// List of page formats sharing this source
  FXObjectListOf<FXTextAnchor> anchs;	// List of anchors (and brackets, being derived).
  FXbool modified_since_create;		// Modified by any widget since creation
  FXbool modified_since_write;		// Modified since last write()
  FXbool modified_since_autosave;	// Modified since last autosave()
  
  FXCharset delimiters;			// Word-splitting delimiters.
  
  FXTextSelection * selection;		// Current text selection.  This points to one
  					// of the anchors in anchs if it exists, otherwise
  					// it is null. 
  FXbool ignoreMarkers;			// Flag to indicate that calls to markRemoved()
  					// and markPlaced() are to be ignored (because the
  					// change is because of a replace() operation).
  FXint last_sel_len;			// Length of most recent selection
  FXchar * last_sel;			// NULL, or text of last selection.  Length of this is
  					// limited to FXTEXTSRC_MAX_LAST_SEL_LEN.
#define FXTEXTSRC_MAX_LAST_SEL_LEN 1024
  FXApp * app;				// The application.  I don't like keeping this here
  					// (IMO it should be a global).

public:
  enum {
    ID_TRACK_TOP,	// Shared widget tracking event: top displayed position changed
    			// Comes with SEL_CHANGED.
    ID_LAST
    };

  long onTrackTop(FXObject*,FXSelector,void*);


protected:
  FXTextSrc();
  
  // Enforce mutual exclusion of anchor visibility
  void enforceMutex(FXTextAnchor * ta);
  
  /// Create a selection bracket
  void createSelection(FXint sp, FXint ep);

  /// Remove an FXTextAnchor from the list of such.  This should only be
  /// called from the TA's destructor (hence the excessive chumminess - I'd rather
  /// declare just the destructor as a friend, but then I'd have to define TAs before
  /// this point.  A weakness of C++).
  friend class FXTextAnchor;
  void removeTextAnchor(const FXTextAnchor * ta);
  
  
private:
  FXTextSrc(const FXTextSrc&);
  FXTextSrc& operator=(const FXTextSrc&);
  
public:
  /// Construct
  FXTextSrc(FXApp * a, FXuint opts);

  /// Test whether constructor completed successfully.  Since ctors may allocate memory etc.,
  /// it is necessary for the caller to check whether construction was sucessful.  If not,
  /// then the dtor should be invoked as soon as possible and an error flagged to the user.
  /// Note that the ctor is exclusively responsible for setting this flag.  The base class
  /// sets it TRUE by default.
  FXbool isOk() const { return ok; }
  
  /// Add another FXPageFormat to the list of such which share this object
  void addPageFormat(FXPageFormat * pf);
  
  /// Remove an FXPageFormat from the list of such which share this object
  void removePageFormat(const FXPageFormat * pf);
  
  /// Add another FXTextAnchor to the list of such.
  void addTextAnchor(FXTextAnchor * ta);
  
  /// Remove an array of FXTextAnchors from the list
  void removeTextAnchors(FXint n, const FXTextAnchor ** tas);
  
  /// Remove all anchors/brackets whose attribute bits specified by mask exactly
  /// match those specified by match.  Returns number removed.
  FXint removeMatchingTextAnchors(FXuint mask, FXuint match);
  
  /// Return number of anchors.
  FXint getNumTextAnchors() const { return anchs.no(); }
  
  /// Find anchors or brackets near given start and/or end points.  Pass either of
  /// sp or ep as -1 to ignore that end in the search.  n is the number of available
  /// elements in the tas array, which is used to return the matching anchor(s).
  /// Return value is number of matching anchors found (up to a maximum of n).
  /// mask and match have same semantics as for removeMatchingTextAnchors().
  FXint findTextAnchors(FXint sp, FXint ep, FXint n, FXTextAnchor ** tas, 
        FXuint mask = 0, FXuint match = 0, FXbool containing = FALSE) const;
  
  /// Return pointer to anchor, given index in range [0..getNumTextAnchors)
  FXTextAnchor * getTextAnchor(FXint idx) const { return anchs[idx]; }
  
  /// Return the smallest bracket completely enclosing a given text range.  If there is
  /// no single "smallest bracket" then the returned bracket is picked arbitrarily.
  /// The candidate brackets are selected based on their response to their isInterested
  /// member function.  True anchors are always excluded since they have zero extent.
  /// May return null if no enclosing interested bracket.
  FXTextAnchor * getEnclosingBracket(FXint startpos, FXint endpos);
  
  /*--- Member functions to be overridden.  In this (base) class, these default to a 'null'
        text source i.e. text is one unmodifiable line of zero characters. ------*/
        
  /// Return a specified text line, given line number (0 is first line).  Return zero-char
  /// string for lines which are 'out-of-range'.  pos is set to the position of the 1st char
  /// of the returned line.  The string contents are only valid until the next call to
  /// this function, since a static string ref is returned.
  virtual const FXString & getLineText(FXuint lineno, FXint & pos) const;
  
  /// Fill a caller-supplied buffer (ptr) of length len with text starting at pos.  The total
  /// number of characters transferred (which may be zero) is returned.  If possible,
  /// a null is written at the end, but this null is not counted in the return value.
  virtual FXint getText(FXint pos, FXchar * ptr, FXint len) const;
  
  /// Set ptr to point to the actual buffer contents (which must not be changed via the
  /// returned pointer).  pos is the desired starting point.  The return value will
  /// indicate the maximum number of contiguous data bytes which may be accessed from
  /// ptr.  On return, pos will be incremented by this amount ready for the next call.
  /// If pos is not valid, then zero is returned and the pointer set to null.
  /// This is the most efficient way of accessing the text contents, however there is
  /// no regard for line breaks since the underlying buffer may not be aware of them.
  virtual FXint getBuf(FXint & pos, const FXchar * & ptr) const; 

  /// Simular to getBuf(), except works back to the start of the buffer.  pos is the desired
  /// endpoint+1.  On return, sets ptr to the start of a contiguous range which includes
  /// at least position (pos-1) and possibly more.  Returns maximum available length, which
  /// may extend to pos or beyond.  pos is updated to the equivalent offset of ptr.
  virtual FXint getBackwardsBuf(FXint & pos, const FXchar * & ptr) const; 

  /// Return position of first character of specified line.
  virtual FXint getPosOfLine(FXuint lineno) const;
  
  /// Return line number containing specified character number.
  virtual FXuint getLineOfPos(FXint pos) const;
  
  /// Return number of lines in text source.
  virtual FXuint lines() const;
  
  /// Return number of characters in text source.  This includes 1 char per newline, but not
  /// a 'trailing null' and not any '\r' characters removed when DOS files are read.  This
  /// is the valid range of the canonical 'character position' metric, not necessarily
  /// the actual number of characters in any file representation.
  virtual FXint length() const; 
  
  /// Return character at given position.  Returns a null if out of range.
  virtual FXchar getChar(FXint pos) const;
  
  /// Interface to replace() below.
  virtual FXint replace(FXint startpos, FXint endpos, 
           const FXString & text, FXbool defer = FALSE, const FXPLText * pltext = NULL);
  
  /// Replace text between [startpos, endpos).  This will invoke change() functions in
  /// all page formats, as well as telling anchors to update themselves if appropriate.
  /// If defer is true, then this is not the last in a sequence of updates which are to
  /// be considered as one 'transaction'.  This flag will propagate to the pager(s) so that
  /// they can defer graphical update for higher efficiency.  All updates in a transaction
  /// group must be performed without passing control back to the GUI.  The transaction
  /// group is terminated by a replace call with defer false.  Returns non-zero if error.
  virtual FXint replace(FXint startpos, FXint endpos, 
           const FXchar * text, FXint n, FXbool defer = FALSE, const FXPLText * pltext = NULL);
  
  /// Undo the most recent num changes.  A change is defined by one or more calls to replace().
  /// Returns the actual number of rollbacks performed.  The maximum number of rollbacks is
  /// determined by the number of changes since the most recent call to write(), revert(),
  /// or creation of this object.  Rollback of the maximum possible number is equivalent to
  /// revert(), 
  virtual FXint rollback(FXint num);
  
  /// Redo num changes that were undone by rollback().  The maximum value for num is the
  /// sum of rollbacks minus the sum of rollforwards which have been performed since the
  /// last call to replace().  Returns the actual number of rollforwards performed.
  virtual FXint rollforward(FXint num);
  
  /// Return position of next rollback (or -1 if no rollback available)
  virtual FXint rollbackPos() const;
  
  /// Return position of next rollforward (or -1 if no rollforward available)
  virtual FXint rollforwardPos() const;
  
  /// Resets the change log to empty.  Returns number of discarded entries.
  virtual FXint resetLog();
  
  /// Write text back to persistent storage, if that makes sense.  Returns 0 if OK, -1 if
  /// operation not supported.  Resets modified_since_write and _autosave flags.  Resets
  /// the change log if successful.  If error (e.g. file system full) then returns a +ve
  /// errno indicator.
  virtual FXint write();
  
  /// Autosave, if that makes sense.  Similar to write(), but intended for periodic call
  /// in order to implement a form of crash protection.  Resets modified_since_autosave flag.
  virtual FXint autosave();
  
  /// Revert to state last saved using write().  It is possible to rollforward, since the
  /// change log is not reset.
  virtual FXint revert();
  
  /// Test whether modifiable (and hence replace(), write(), rollback() etc. are likely to
  /// have a meaningful effect).  This may return false, yet replace() works, in the case
  /// that the permanent storage e.g. file is not writable by this user.
  virtual FXbool isModifiable() const; 
  
  /// Search text using the parameters in fd.  pos is starting position, epos is last allowable
  /// position+1.  epos may be <= pos, in which case wrap-around is allowed if not found
  /// before end of text.  Returns position found, or -1 if not found.  len is set to length
  /// of match.  pos may be -ve, in which case it is wrapped to last character; or > length(), in
  /// which case it is wrapped to 0.  The virtual newline which is the last text character
  /// participates in the search.  pos and epos represent the allowable position of the
  /// first char of the matching string for forward searches, or the last character for
  /// backward searches.  If widget is non-null, and an error is detected such as a malformed
  /// regular expression, then a message box is popped up, and the return code will be -2.
  virtual FXint find(FXFindData * fd, FXint pos, FXint epos, 
                     FXint & len, FXWindow * widget = NULL) const;

  /// Specialised version of find(), for literal searches.
  virtual FXint findLiteralForward(FXFindData * fd, FXint pos, FXint epos, 
          const FXchar * ss, FXint sslen, FXbool case_sens, FXbool whole_word) const; 
  
  /// Specialised version of find(), for literal searches.
  virtual FXint findLiteralBackward(FXFindData * fd, FXint pos, FXint epos, 
          const FXchar * ss, FXint sslen, FXbool case_sens, FXbool whole_word) const;
  
  /// Reset the position of the output/output2 brackets to EOF.  This is typically called
  /// when the input bracket gets enter key from user.
  virtual void resetOutput();
  
  virtual FXProcess * getProcess() const { return NULL; }
  /*--- end of virtual members ---*/
  
  /// Return end of previous word, plus 1.
  FXint leftWord(FXint pos) const; 
  
  /// Return start of next word.
  FXint rightWord(FXint pos) const;
  
  /// Return start of this word.
  FXint wordStart(FXint pos) const;
  
  /// Return end of this word, plus 1. 
  FXint wordEnd(FXint pos) const;
  
  /// Return start of line containing pos.
  FXint lineStart(FXint pos) const { return getPosOfLine(getLineOfPos(pos)); }
  
  /// Return end of line containing pos.
  FXint lineEnd(FXint pos) const { return getPosOfLine(getLineOfPos(pos)+1)-1; }
  
  /// Return start of next line.
  FXint nextLine(FXint pos) const { return getPosOfLine(getLineOfPos(pos)+1); }
  
  /// Return start of previous line.
  FXint prevLine(FXint pos) const { return getPosOfLine(getLineOfPos(pos)-1); }
  
  /// Return modification state.  Parameter is 0, 1, or 2 for since create, write or
  /// autosave respectively.
  virtual FXbool isModified(FXuint which = 1) const;
  
  /// Set modification state.  Parameter is 0, 1, or 2 for since create, write or
  /// autosave respectively.
  void setModified(FXbool mod, FXuint which = 1);
  
  /// Iterate through list of markers (anchors and brackets) in order for them to update
  /// the given page format.  The specified range of positions (sp to ep+1) allows the
  /// markers to optimise the update process.  These positions represent the first and
  /// last chars of interest.  This is called from FXPageFormat::refresh()
  /// when it needs to reconstruct its attribute matrix.  In turn, this calls the tellAttrs()
  /// method of all anchors. 
  void enumerateMarkers(FXPageFormat * pf, FXint startpos, FXint endpos) const;
  
  /// Broadcast fact that given anchor (or bracket) has been removed.
  /// If pf is null, all page formats are notified.  Otherwise, only the specific
  /// page format is notified.  If temp is true, graphical update to be deferred.
  /// sp and/or ep (if not defaulted) specify a sub-range to act upon; otherwise
  /// the entire bracket range is assumed.
  void markRemoved(FXTextAnchor * ta, FXPageFormat * pf, FXbool temp,
  		   FXint sp = -1, FXint ep = -1);
  
  /// Broadcast fact that given anchor (or bracket) has been placed at the given
  /// position (sp to ep).
  /// If pf is null, all page formats are notified.  Otherwise, only the specific
  /// page format is notified.  If temp is true, graphical update to be deferred.
  void markPlaced(FXTextAnchor * ta, FXPageFormat * pf, FXbool temp,
  		   FXint sp = -1, FXint ep = -1);
  
  /// Return delimiter char set.
  const FXCharset & getDelimiters() const { return delimiters; }
  
  /// Set new delimiters.
  void setDelimiters(const FXCharset & delims) { delimiters = delims; }

  /// Remove the current selection
  void killSelection();
  
  /// Set or update selection positions
  void setSelection(FXint sp, FXint ep);
  
  /// Return start of selection (-1 if none)
  FXint getSelectionStart() const;
  
  /// Return end of selection (-1 if none)
  FXint getSelectionEnd() const;
  
  /// Return 1 if there is a valid selection, else 0.  sp and ep are updated to current
  /// start and end.
  FXint getSelection(FXint & sp, FXint & ep) const;
  
  /// Tell if there is a valid selection
  FXbool isSelection() const;
  
  /// Return pointer to buffer containing most recent selected text, length in len.
  FXchar * lastSelection(FXint & len) const { len = last_sel_len; return len ? last_sel : NULL; }
  
  /// Return application.
  FXApp * getApp() const { return app; }

  /// Save to a stream
  virtual void save(FXStream& store) const;

  /// Load from a stream
  virtual void load(FXStream& store);

  /// Destructor.
  virtual ~FXTextSrc();
  };
  

class FXTextBuf;
class FXTextBufP;
class FXTextBufS;
class FXTextMap;


/// A text source without secondary (i.e. file system) storage. 
class FXAPI FXMemTextSrc : public FXTextSrc {
  FXDECLARE(FXMemTextSrc)
protected:  
  FXTextBuf * textbuf;	// Text buffer
  FXTextMap * map;	// Line/position mapping
  FXbool init;		// True if being initialised.
  
  // Used for rollback/forward.  This really is supposed to be protected (and gcc allows it to be),
  // but Win32 compilers complain unless it is public.
public:
  struct Transaction {
    FXint pos;		// Position of first char replaced
    FXint olen;		// Length of chars that were replaced
    FXint nlen;		// Length of replacement
    FXbool notlast;	// True if not last in transaction group (from defer flag in replace())
    FXchar * before;	// Before-image (length olen); start of buffer memory
    FXchar * after;	// Replacement data (length nlen), points to before+olen.
    };

protected:
  FXArray<Transaction> log;	// Change log 
  FXbool rolling;	// True if currently in rollback/forward mode.  If true, prevents
  			// recursive logging of updates.
  FXint  logent;	// Index of next available log entry.  If this equals log.no(), then
  			// there is no rollforward available, and any changes will be appended
  			// to the log.  If less than log.no(), then log.no()-logent rollforward
  			// entries are available.  modified_since_write will be false if and
  			// only if logent == 0. 
  
  FXMemTextSrc();
  FXuint getLineOfPos(FXint pos, FXuint & node, FXuint & Ln, FXuint & Pn) const;
  Transaction * truncateLog(FXint num);
  
private:
  FXMemTextSrc(const FXMemTextSrc&);
  FXMemTextSrc& operator=(const FXMemTextSrc&);
  
public:
  /// Construct with 'empty' string
  FXMemTextSrc(FXApp * a, FXuint opts);
  
  /// Construct with initial contents.
  FXMemTextSrc(FXApp * a, FXuint opts, const FXchar * text, FXint textlen);
  
  virtual const FXString & getLineText(FXuint lineno, FXint & pos) const;
  virtual FXint getText(FXint pos, FXchar * ptr, FXint len) const;
  virtual FXint getBuf(FXint & pos, const FXchar * & ptr) const; 
  virtual FXint getBackwardsBuf(FXint & pos, const FXchar * & ptr) const; 
  virtual FXint getPosOfLine(FXuint lineno) const;
  virtual FXuint getLineOfPos(FXint pos) const;
  virtual FXuint lines() const;
  virtual FXint length() const;
  virtual FXchar getChar(FXint pos) const;
  virtual FXint replace(FXint startpos, FXint endpos, const FXString & text, 
                 FXbool defer = FALSE, const FXPLText * pltext = NULL);
  virtual FXint replace(FXint startpos, FXint endpos, const FXchar * text, FXint n, 
                 FXbool defer = FALSE, const FXPLText * pltext = NULL);
  virtual FXint rollback(FXint num);
  virtual FXint rollforward(FXint num);
  virtual FXint rollbackPos() const;
  virtual FXint rollforwardPos() const; 
  virtual FXint resetLog();
  virtual FXint write();
  virtual FXint autosave();
  virtual FXint revert();
  virtual FXbool isModifiable() const;
  
  
  /// Save to a stream
  virtual void save(FXStream& store) const;

  /// Load from a stream
  virtual void load(FXStream& store);

  /// Destructor.
  virtual ~FXMemTextSrc();
  };

/// A text source which uses the file system.
class FXAPI FXFileTextSrc : public FXMemTextSrc {
  FXDECLARE(FXFileTextSrc)
public:

  /// Describes the external (file system) representation of the text, mainly with regard
  /// to line ending conventions.
  enum le_mode {
    AUTO,		/// Automagically guess (pass to ctor, open() etc.)
    LF,			/// '\n' as per Unix
    UNIX = LF,
    LF_ATEOF,		/// '\n' style, and ensure at least one '\n' at EOF.
    CRLF,		/// '\r\n' as per DOS etc.
    DOS = CRLF,
    CRLF_ATEOF,		/// '\r\n' style, and ensure at least one '\r\n' at EOF.
    BINARY		/// No line endings.  This type cannot be saved, because fake
    			/// newlines are added to break up potentially long lines,
    			/// and the fake newlines cannot be distinguished from real ones.
    };
  
  enum {
    /// SIGCHLD signal received from shell (if openShell() called)
    ID_SIGCHLD = FXTextSrc::ID_LAST, 
    ID_LAST
    };

#ifndef WIN32  
  // SEL_SIGNAL/ID_SIGCHLD message
  long onSigSIGCHLD(FXObject*,FXSelector,void*);
#endif

protected:
  FXString file;	// File name of current file, or "" if none.
  FXString logfile;	// File name of transaction log file, or "" if logging to memory.
  FXString autosavefile;// File name of autosave file, or "" if none.
  le_mode mode;		// Line-ending convention.
  FXbool lock;		// True if given file is locked against concurrent update (at least
  			// by the same FOX application, or even other applications if the
  			// operating system supports it).
  
  FXbool anon;		// True if file name not specified (i.e. file == "").  Other files
  			// may be valid, but this flag indicates a name needs to be chosen
  			// before this text can be saved.
  static FXint uniq;	// File counter for generating anonymous autosave names.
  FXint  anon_num;	// Anonymous file number for this file.
  FXbool as_hold;	// Do not autosave, because currently prompting user via dialogue.
  const FXchar ** patterns;	// File name pattern(s) for open file dialogue.
  FXProcess * proc;	// Shell process
  FXOutputBracket * output, * output2;	// stdout and stderr
  FXInputBracket * input;		// stdin
  FXbool use_pipe;	// Whether to be a terminal (false) or just look like a non-interactive
  			// socket/pipe (true).
  FXString resource;	// Resource name part
  FXString protocol;	// Protocol name part e.g. shell://
  FXbool is_shell;	// True if this is shell text source (else file)

  FXint writeTo(const FXString & fname);
  FXFileTextSrc();
  // Parse the given name and set resource and isShell members.
  void getProtocol(const FXString & name);
  
  
private:
  FXFileTextSrc(const FXFileTextSrc&);
  FXFileTextSrc& operator=(const FXFileTextSrc&);
  
public:
  /// Construct with unnamed file
  FXFileTextSrc(FXApp * a, FXuint opts);
  
  /// Construct with initial contents provided by file 'name'.  Other files are optional,
  /// and specify the transaction log and/or autosave file names.  m specifies a line-ending
  /// convention (the default attempts to guess it).  lck specifies whether to lock the
  /// file against concurrent update.  The file is not actually opened at this time; the
  /// caller must call open(FXWindow *, FXbool).
  FXFileTextSrc(FXApp * a, FXuint opts, const FXString & name, le_mode m = AUTO, FXbool lck = TRUE);
  
  //virtual FXint rollback(FXint num);
  //virtual FXint rollforward(FXint num);
  //virtual FXint rollbackPos() const;
  //virtual FXint rollforwardPos() const;  
  virtual FXint write();
  virtual FXint autosave();
  virtual FXint revert();
  virtual FXbool isModifiable() const; 
  virtual FXbool isModified(FXuint which = 1) const;
  virtual void resetOutput();
  
  // Additional functions for files.  These correspond quite closely with the usual
  // file menu semantics.
  
  /// Generate an autosave file name given a file name to open.  In this class, the implementation
  /// assumes that the extension '.jic' may be added to the given file name.  This may not
  /// be satisfactory for all file systems (e.g. old DOS FAT) so it may be overridden in
  /// derived classes.  The most important rule for generating names is that they should map
  /// 1:1 to 'user' file names, because when a new file is opened, this class checks for a
  /// corresponding autosave file name with modification time more recent than the user file.
  /// In the case that the current file is untitled (i.e. isNamed() returns false), then
  /// names of the form 'untitled_PID_NNN.jic' are generated where PID is the Unix PID (or
  /// some other unique process number) and NNN is a sequence number incremented within
  /// this process.  On recovery from a crash, the user will need to search for files ending
  /// with '.jic' if he was editing untitled files at the time.
  virtual FXString makeAutosaveFileName(const FXString & name);
  
  /// Generate a log file name given file name to open.  Works as for makeAutosaveFileName()
  /// except that the extension is '.log'.  Note that this function should be called after
  /// makeAutosaveFileName() in order to use the same unique anonymous file number.
  virtual FXString makeLogFileName(const FXString & name);
  
  /// Return whether a 'current file' is assigned.
  FXbool isNamed() const { return !anon; }
  
  /// Open new file.  Parms same as for ctor.  The previous file is saved if: prompt is true,
  /// it has been modified since write(), and user answers 'save' to message box.  Otherwise,
  /// the previous file is not saved.  The user can press 'cancel' in which case -1 is returned
  /// and no other action taken.  Otherwise, returns 0.  If an error occurs during the save,
  /// a message box is displayed regardless of whether prompt is true, and a positive errno
  /// indicator is returned.  If an error occurs in the load (e.g. non-existent file) then
  /// the message box will indicate the error, and the contents will be unchanged.  If name
  /// is empty string, then user will be prompted for file name.  If the file name is
  /// empty from the dialogue, then neo() is called. 
  /// 'widget' must be an appropriate owner for the message boxes and/or dialogues.
  FXint open(FXWindow * widget, FXbool prompt, const FXString & name,
  	     le_mode m = AUTO, FXbool lck = TRUE);

#ifndef WIN32
  /// Open Unix shell process.  Turns the text source into a command-line tool.
  /// Returns non-zero if error as for open().
  FXint openShell(FXWindow * widget, FXbool prompt, const FXchar * rsrc);
#endif

  /// Ask whether this is connected to a shell
  FXbool isShell() const { return is_shell; }

  /// Open current file (specified to ctor).
  FXint open(FXWindow * widget, FXbool prompt);
  
  /// Open unnamed, empty file.  Semantics otherwise the same as for open().  Caller should
  /// call setMode() before writing unless the default (LF or CRLF depending on platform) is
  /// appropriate.  Really wanted to call this function 'new', but that's a C++ operator...
  /// Returns -1 if the operation was cancelled because the user cancelled from the save-existing
  /// message box.
  FXint neo(FXWindow * widget, FXbool prompt);
  
  /// Interface to write(), with user-friendly error handling.
  FXint save(FXWindow * widget, FXbool prompt);

  /// Allow saving of current file if modified.  Call this before destructor.
  FXint maybeSave(FXWindow * widget, FXbool prompt);
  
  /// Save as new file.  Also changes log, autosave, mode and lock.  Semantics are equivalent
  /// to copying current text to the new file, then opening that file.  If prompt is true,
  /// and new file already exists, then message box will confirm overwrite.  If overwrite
  /// is denied, then returns -1 with no further action.  If name is "", then user will
  /// be prompted for save file name. 
  FXint saveAs(FXWindow * widget, FXbool prompt, const FXString & name,
  	       le_mode m = AUTO, FXbool lck = TRUE);

  /// Show error message box with analysis of error code returned from write(), autosave().
  void showError(FXWindow * widget, FXint err, 
                 const FXchar * heading, const FXchar * context); 

  /// Return current file name, or "untitled" if isNamed() returns false.
  const FXString & getFile() const { return file; }
    
  /// Return current transaction log file name, or "" if none.
  const FXString & getLogFile() const { return logfile; }
    
  /// Return current autosave file name, or "" if none.
  const FXString & getAutosaveFile() const { return autosavefile; }
  
  /// Return current line-ending mode
  le_mode getMode() const { return mode; }
  
  /// Set line-ending mode
  void setMode(le_mode m);
  
  /// Return current lock status
  FXbool getLock() const { return lock; }
  
  /// Set lock status.  Returns True if obtained lock, else False if someone else has it.
  FXbool setLock(FXbool lck = TRUE);
  
  /// Set file open dialogue pattern(s)
  void setPatterns(const FXchar ** patt) { patterns = patt; }
  
  /// Get pattern
  const FXchar ** getPatterns() const { return patterns; }
  
  /// Get process
  virtual FXProcess * getProcess() const { return proc; }
  
  /// Save to a stream
  virtual void save(FXStream& store) const;

  /// Load from a stream
  virtual void load(FXStream& store);

  /// Destructor.
  virtual ~FXFileTextSrc();
  };

#endif